#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <inttypes.h>
#include <arpa/inet.h>
#ifndef VARNISH_VERSION_BEFORE_410
#include <errno.h>
#endif
#ifndef VARNISH_VERSION_BEFORE_600
#include <unistd.h>
#endif

#ifdef VARNISH_VERSION_BEFORE_600
#include "vcl.h"
#include "vrt.h"
#ifdef VARNISH_VERSION_BEFORE_400
#include "vrt_obj.h"

#include "bin/varnishd/cache.h"
#else
#include "vsa.h"

#include "cache/cache.h"
#endif
#else // 600+
#include "cache/cache_varnishd.h"
#endif
#ifdef VARNISH_VERSION_BEFORE_410
#ifdef VARNISH_VERSION_BEFORE_400
#include "bin/varnishd/cache_backend.h"
#else
#include "cache/cache_backend.h"
#endif
#else
#include "cache/cache_director.h"
#endif
#ifndef VARNISH_VERSION_BEFORE_500
#include "cache/cache_filter.h"
#endif

#ifndef VARNISH_VERSION_BEFORE_600
#include "vre.h"
#endif

#include "base64.h"
#include "vcc_if.h"
#include "config.h"

#define MAX_PARAM_DATALEN (10 * 1024)

#ifdef VARNISH_VERSION_BEFORE_400
#define VRT_CTX_MAGIC SESS_MAGIC
#define VRT_CTX struct sess *ctx
#define VCL_VOID void
#define VCL_STRING const char *
#define VCL_BOOL unsigned
#define VCL_BACKEND struct director *
#define HTTP_HDR_METHOD HTTP_HDR_REQ

#define CTX_REQ ctx
#define CTX_HTTP_REQ ctx->http

#define LOG(ctx, ...)                           \
  WSP(ctx, __VA_ARGS__)

#define HTTP_IS_REQ_HDR(ctx, ...)               \
  http_HdrIs(ctx->http, __VA_ARGS__)
#define HTTP_GET_REQ_HDR(ctx, ...)              \
  http_GetHdr(ctx->http, __VA_ARGS__)
#define HTTP_SET_REQ_HDR(ctx, ...)                                    \
  http_SetHeader(ctx->wrk, ctx->wrk->htc->fd, ctx->http, __VA_ARGS__)
#define HTTP_PRINTF_REQ_HDR(ctx, ...)                                    \
  http_PrintfHeader(ctx->wrk, ctx->wrk->htc->fd, ctx->http, __VA_ARGS__)

#define HTTP_GET_RESP_HDR(ctx, ...)             \
  http_GetHdr(ctx->wrk->resp, __VA_ARGS__)
#define HTTP_SET_RESP_HDR(ctx, ...)                                     \
  http_SetHeader(ctx->wrk, ctx->wrk->htc->fd, ctx->wrk->resp, __VA_ARGS__)

#else

#ifdef VARNISH_VERSION_BEFORE_410
#define VRT_CTX const struct vrt_ctx *ctx
#endif

#define CTX_REQ ctx->req
#define CTX_HTTP_REQ ctx->http_req

#define LOG(ctx, ...)                           \
  VSLb(ctx->vsl,  __VA_ARGS__)

#define HTTP_IS_REQ_HDR(ctx, ...)               \
  http_HdrIs(ctx->http_req, __VA_ARGS__)
#define HTTP_GET_REQ_HDR(ctx, ...)              \
  http_GetHdr(ctx->http_req, __VA_ARGS__)
#define HTTP_PRINTF_REQ_HDR(ctx, ...)           \
  http_PrintfHeader(ctx->http_req, __VA_ARGS__)
#define HTTP_SET_REQ_HDR(ctx, ...)              \
  http_SetHeader(ctx->http_req, __VA_ARGS__)

#define HTTP_GET_RESP_HDR(ctx, ...)             \
  http_GetHdr(ctx->http_resp, __VA_ARGS__)
#define HTTP_SET_RESP_HDR(ctx, ...)             \
  http_SetHeader(ctx->http_resp, __VA_ARGS__)

#endif

#ifdef VARNISH_VERSION_BEFORE_410
struct vmod_storage;
#endif

#ifdef VARNISH_VERSION_BEFORE_630
#define WS_ReserveAll(ws) WS_Reserve(ws, 0)
#endif


struct vmod_data_dome_shield_cfg {
  unsigned				 magic;
#define VMOD_DATA_DOME_SHIELD_CFG_MAGIC  0x77aaff66

  VCL_BACKEND				 backend;
  char                                  *key;

  char                                  *re_raw;
  vre_t                                 *re;

  char                                  *re_exclusion_raw;
  vre_t                                 *re_exclusion;

  char                                  *cookie;

  char                                  *protocol;

  char                                   host_name[128];

  txt                                    validate_url;
  txt                                    validate_method;
  txt                                    validate_proto;

  unsigned                               send_user_agent;
  unsigned                               send_ip;
  unsigned                               send_port;
  unsigned                               send_ClientID;
  unsigned                               send_host;
  unsigned                               send_referer;
  unsigned                               send_request;
  unsigned                               send_protocol;
  unsigned                               send_method;
  unsigned                               send_cookies;
  unsigned                               send_cookies_len;
  unsigned                               send_time_request;
  unsigned                               send_server_hostname;
  unsigned                               send_module_name;
  unsigned                               send_module_version;
  unsigned                               send_post_param_len;
  unsigned                               send_server_name;
  unsigned                               send_x_forward_for_ip;
  unsigned                               send_headers_list;
  unsigned                               send_authorization_len;
  unsigned                               send_x_requested_with;
  unsigned                               send_origin;
  unsigned                               send_connection;
  unsigned                               send_pragma;
  unsigned                               send_cache_control;
  unsigned                               send_content_type;
  unsigned                               send_from;
  unsigned                               send_x_real_ip;
  unsigned                               send_via;
  unsigned                               send_true_client_ip;
  unsigned                               send_accept;
  unsigned                               send_accept_charset;
  unsigned                               send_accept_encoding;
  unsigned                               send_accept_language;

  unsigned                               is_valid_force_close_banned_request;

#define HTTPH(a, b) char *b;
#include "http_extra_headers.h"
#undef HTTPH

#ifdef VARNISH_VERSION_BEFORE_410
  struct vmod_storage                   *storage;
  pthread_mutex_t                        storage_mutex;
#endif
};

struct vmod_data_dome_shield_client_id {
  char         *client_id;
};

struct vmod_data_dome_shield_status {
  unsigned            magic;
#define VMOD_DATA_DOME_SHIELD_STATUS_MAGIC  0x77aaff77

#ifdef VARNISH_VERSION_BEFORE_410
#ifdef VARNISH_VERSION_BEFORE_400
  struct sess                              *owner;
#else // 400
  struct req                               *owner;
#endif
  VTAILQ_ENTRY(vmod_data_dome_shield_status) list;
#else // 410
  struct vmod_priv *priv_task;
#endif

  struct vmod_data_dome_shield_client_id    client_id;

  VCL_BACKEND           orign_backend;
#ifdef VARNISH_VERSION_BEFORE_640
#ifndef VARNISH_VERSION_BEFORE_400
  volatile enum req_body_state_e      orign_req_body_status;
#endif
  enum body_status          orign_body_status;
#else
  body_status_t                       orign_req_body_status;
  body_status_t                       orign_body_status;
#endif
  char                                     *orign_pipeline_b;
  char                                     *orign_pipeline_e;
  char                                     *orign_rxbuf_b;
  char                                     *orign_rxbuf_e;
#ifdef VARNISH_VERSION_BEFORE_400
  uint16_t                                  orign_shd;
  txt                                      *orign_hd;
  unsigned char                            *orign_hdf;
  uint16_t                                  orign_nhd;
#else
  ssize_t                                   orign_content_length;

  struct http                              *orign_http_req;
  struct http                              *orign_req_http;
#endif
#ifndef VARNISH_VERSION_BEFORE_500
  struct vfp_ctx                            orign_vfc[1];
#endif

  unsigned int                              headers_restored;

  struct timeval tv;

  char                                      body[MAX_PARAM_DATALEN];
  ssize_t                                   body_len;
};

#ifdef VARNISH_VERSION_BEFORE_410

struct vmod_storage {
  unsigned magic;
#define VMOD_STORAGE_MAGIC 0x4EE5FB2E

  VTAILQ_HEAD(, vmod_data_dome_shield_status) statuses;
};


static void remove_status(VRT_CTX, struct vmod_data_dome_shield_cfg *cfg, struct vmod_data_dome_shield_status *status) {
#ifdef VARNISH_VERSION_BEFORE_400
  LOG(ctx, SLT_Debug, "DataDome remove status for client: %s:%s", ctx->addr, ctx->port);
#else
  VSLb(ctx->vsl, SLT_Debug, "DataDome remove status for client: %s:%s",
   ctx->req->sp->client_addr_str, ctx->req->sp->client_port_str);
#endif

  pthread_mutex_lock(&cfg->storage_mutex);
  VTAILQ_REMOVE(&cfg->storage->statuses, status, list);
  pthread_mutex_unlock(&cfg->storage_mutex);
}


static void
set_status(VRT_CTX, struct vmod_data_dome_shield_cfg *cfg, struct vmod_data_dome_shield_status *status) {
#ifdef VARNISH_VERSION_BEFORE_400
  LOG(ctx, SLT_Debug, "DataDome set status for client: %s:%s", ctx->addr, ctx->port);
#else
  VSLb(ctx->vsl, SLT_Debug, "DataDome set status for client: %s:%s",
   ctx->req->sp->client_addr_str, ctx->req->sp->client_port_str);
#endif

  status->owner = CTX_REQ;

  pthread_mutex_lock(&cfg->storage_mutex);
  VTAILQ_INSERT_TAIL(&cfg->storage->statuses, status, list);
  pthread_mutex_unlock(&cfg->storage_mutex);
}


static struct vmod_data_dome_shield_status *
get_status(VRT_CTX, struct vmod_data_dome_shield_cfg *cfg) {
#ifdef VARNISH_VERSION_BEFORE_400
  LOG(ctx, SLT_Debug, "DataDome get status for client: %s:%s", ctx->addr, ctx->port);
#else
  VSLb(ctx->vsl, SLT_Debug, "DataDome get status for client: %s:%s",
   ctx->req->sp->client_addr_str, ctx->req->sp->client_port_str);
#endif

  struct vmod_data_dome_shield_status *status;

  pthread_mutex_lock(&cfg->storage_mutex);
  VTAILQ_FOREACH(status, &cfg->storage->statuses, list) {
    CHECK_OBJ_NOTNULL(status, VMOD_DATA_DOME_SHIELD_STATUS_MAGIC);
    if (status->owner == CTX_REQ) {
      pthread_mutex_unlock(&cfg->storage_mutex);
      return status;
    }
  }
  pthread_mutex_unlock(&cfg->storage_mutex);

  return NULL;
}
#endif

VCL_VOID vmod_cfg_free(void *ptr)
{
  AN(ptr);

  struct vmod_data_dome_shield_cfg *config = ptr;

  CHECK_OBJ_NOTNULL(config, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  if (config->key != NULL) {
    free(config->key);
    config->key = NULL;
  }

  if (config->cookie != NULL) {
    free(config->cookie);
    config->cookie = NULL;
  }

  if (config->protocol != NULL) {
    free(config->protocol);
    config->protocol = NULL;
  }

  if (config->re_raw != NULL) {
    free(config->re_raw);
    config->re_raw = NULL;
  }

  if (config->re_exclusion_raw != NULL) {
    free(config->re_exclusion_raw);
    config->re_exclusion_raw = NULL;
  }

  if (config->re != NULL) {
    VRT_re_fini(config->re);
    config->re = NULL;
  }

  if (config->re_exclusion != NULL) {
    VRT_re_fini(config->re_exclusion);
    config->re_exclusion = NULL;
  }

#ifdef VARNISH_VERSION_BEFORE_410
  FREE_OBJ(config->storage);
#endif

#define HTTPH(a, b) free(config->b);
#include "http_extra_headers.h"
#undef HTTPH

  FREE_OBJ(config);
}


VCL_VOID
vmod_init(VRT_CTX, struct vmod_priv *priv_top, VCL_BACKEND backend, VCL_STRING key)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif
#ifndef VARNISH_VERSION_BEFORE_410
  AN(ctx->msg);
#endif
  AN(priv_top);
  AZ(priv_top->priv);

  struct vmod_data_dome_shield_cfg *cfg;

  ALLOC_OBJ(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);
  AN(cfg);

#define HTTPH(a, b) {                     \
    cfg->b = strdup("*" a ":");           \
    AN(cfg->b);                           \
    cfg->b[0] = sizeof(a);                \
}
#include "http_extra_headers.h"
#undef HTTPH
  priv_top->priv = cfg;
  priv_top->free = vmod_cfg_free;

#ifdef VARNISH_VERSION_BEFORE_410
  ALLOC_OBJ(cfg->storage, VMOD_STORAGE_MAGIC);
  AN(cfg->storage);
  VTAILQ_INIT(&cfg->storage->statuses);

  AZ(pthread_mutex_init(&cfg->storage_mutex, NULL));
#endif

  AN(backend);
  cfg->backend = backend;

  cfg->key = strdup(key);
  AN(cfg->key);

  cfg->cookie = strdup("datadome");
  AN(cfg->cookie);

  cfg->protocol = strdup("http");
  AN(cfg->protocol);

  cfg->validate_url.b = "/validate-request/";
  cfg->validate_url.e =
    cfg->validate_url.b + sizeof("/validate-request/") - 1;

  cfg->validate_method.b = "POST";
  cfg->validate_method.e =
    cfg->validate_method.b + sizeof("POST") - 1;

  cfg->validate_proto.b = "HTTP/1.1";
  cfg->validate_proto.e =
    cfg->validate_proto.b + sizeof("HTTP/1.1") - 1;

  cfg->send_user_agent = 1;
  cfg->send_ip = 1;
  cfg->send_port = 1;
  cfg->send_ClientID = 1;
  cfg->send_host = 1;
  cfg->send_referer = 1;
  cfg->send_request = 1;
  cfg->send_protocol = 1;
  cfg->send_method = 1;
  cfg->send_cookies_len = 1;
  cfg->send_time_request = 1;
  cfg->send_server_hostname = 1;
  cfg->send_module_name = 1;
  cfg->send_module_version = 1;
  cfg->send_post_param_len = 1;
  cfg->send_server_name = 1;
  cfg->send_x_forward_for_ip = 1;
  cfg->send_headers_list = 1;
  cfg->send_authorization_len = 1;
  cfg->send_x_requested_with = 1;
  cfg->send_origin = 1;
  cfg->send_connection = 1;
  cfg->send_pragma = 1;
  cfg->send_cache_control = 1;
  cfg->send_content_type = 1;
  cfg->send_from = 1;
  cfg->send_x_real_ip = 1;
  cfg->send_via = 1;
  cfg->send_true_client_ip = 1;
  cfg->send_accept = 1;
  cfg->send_accept_charset = 1;
  cfg->send_accept_encoding = 1;
  cfg->send_accept_language = 1;

  cfg->is_valid_force_close_banned_request = 1;

  if (gethostname(cfg->host_name, sizeof(cfg->host_name)) < 0) {
    cfg->host_name[0] = '\0';
  }

  cfg->re = NULL;
  cfg->re_raw = strdup("");
  AN(cfg->re_raw);

  cfg->re_exclusion_raw = strdup("\\.(avi|flv|mka|mkv|mov|mp4|mpeg|mpg|mp3|flac|ogg|ogm|opus|wav|webm|webp|bmp|gif|ico|jpeg|jpg|png|svg|svgz|swf|eot|otf|ttf|woff|woff2|css|less|js)$");
  AN(cfg->re_exclusion_raw);

  vre_t *t;
  const char *error;
  int erroroffset;

  t = VRE_compile(cfg->re_exclusion_raw, VRE_CASELESS, &error, &erroroffset);
  AN(t);
  cfg->re_exclusion = t;
}


VCL_VOID vmod_validate_url(VRT_CTX, struct vmod_priv *priv_top, VCL_STRING value) {
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_410
  cfg->validate_url.b = (char *)value;
  cfg->validate_url.e = (char *)value + strlen(value);
#else
  cfg->validate_url.b = value;
  cfg->validate_url.e = value + strlen(value);
#endif
}


VCL_VOID vmod_uri_regex(VRT_CTX, struct vmod_priv *priv_top, VCL_STRING pattern)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  if (cfg->re) {
      free(cfg->re_raw);
      VRT_re_fini(cfg->re);
  }

  cfg->re_raw = strdup(pattern);
  AN(cfg->re_raw);

  vre_t *t;
  const char *error;
  int erroroffset;

  t = VRE_compile(cfg->re_raw, VRE_CASELESS, &error, &erroroffset);
  AN(t);
  cfg->re = t;
}


VCL_VOID vmod_uri_regex_exclusion(VRT_CTX, struct vmod_priv *priv_top, VCL_STRING pattern)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  if (cfg->re_exclusion) {
      free(cfg->re_exclusion_raw);
      VRT_re_fini(cfg->re_exclusion);
  }

  cfg->re_exclusion_raw = strdup(pattern);
  AN(cfg->re_exclusion_raw);

  vre_t *t;
  const char *error;
  int erroroffset;

  t = VRE_compile(cfg->re_exclusion_raw, VRE_CASELESS, &error, &erroroffset);
  AN(t);
  cfg->re_exclusion = t;
}


VCL_VOID vmod_cookie(VRT_CTX, struct vmod_priv *priv_top, VCL_STRING name)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  free(cfg->cookie);

  cfg->cookie = strdup(name);
  AN(cfg->cookie);
}


VCL_VOID vmod_debug_params(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);
}


VCL_VOID vmod_send_user_agent(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_user_agent = value;
}


VCL_VOID vmod_send_ip(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_ip = value;
}


VCL_VOID vmod_send_port(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_port = value;
}


VCL_VOID vmod_send_client_id(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_ClientID = value;
}


VCL_VOID vmod_send_host(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_host = value;
}


VCL_VOID vmod_send_referer(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_referer = value;
}


VCL_VOID vmod_send_request(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_request = value;
}


VCL_VOID vmod_send_protocol(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value, VCL_STRING protocol)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_protocol = value;
  free(cfg->protocol);
  cfg->protocol = strdup(protocol);
  AN(cfg->protocol);
}


VCL_VOID vmod_send_method(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_method = value;
}


VCL_VOID vmod_send_cookies_len(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_cookies_len = value;
}


VCL_VOID vmod_send_time(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_time_request = value;
}


VCL_VOID vmod_send_server_hostname(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_server_hostname = value;
}


VCL_VOID vmod_send_module_name(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_module_name = value;
}


VCL_VOID vmod_send_module_version(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_module_version = value;
}


VCL_VOID vmod_send_post_param_len(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_post_param_len = value;
}


VCL_VOID vmod_send_server_name(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_server_name = value;
}


VCL_VOID vmod_send_x_forward_for_ip(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_x_forward_for_ip = value;
}


VCL_VOID vmod_send_headers_list(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_headers_list = value;
}


VCL_VOID vmod_send_authorization_len(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_authorization_len = value;
}


VCL_VOID vmod_send_x_requested_with(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_x_requested_with = value;
}


VCL_VOID vmod_send_origin(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_origin = value;
}


void vmod_send_connection(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_connection = value;
}


void vmod_send_pragma(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_pragma = value;
}


void vmod_send_cache_control(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_cache_control = value;
}


void vmod_send_content_type(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_content_type = value;
}


void vmod_send_from(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_from = value;
}


void vmod_send_x_real_ip(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_x_real_ip = value;
}


void vmod_send_via(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_via = value;
}


void vmod_send_true_client_ip(VRT_CTX, struct vmod_priv *priv_top, unsigned value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_true_client_ip = value;
}


VCL_VOID vmod_send_accept(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_accept = value;
}


VCL_VOID vmod_send_accept_charset(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_accept_charset = value;
}


VCL_VOID vmod_send_accept_encoding(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_accept_encoding = value;
}


VCL_VOID vmod_send_accept_language(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
#ifdef VARNISH_VERSION_BEFORE_400
  AZ(ctx);
#else
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
#endif

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  cfg->send_accept_language = value;
}


VCL_BOOL vmod_is_regex_matched(VRT_CTX, struct vmod_priv *priv_top)
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_400
  ssize_t uri_len = ctx->http->hd[HTTP_HDR_URL].e - ctx->http->hd[HTTP_HDR_URL].b;

  const char *question = (const char *)memchr(ctx->http->hd[HTTP_HDR_URL].b, '?', uri_len);
  if (question) {
    uri_len = question - ctx->http->hd[HTTP_HDR_URL].b;
  }

  const char *uri = ctx->http->hd[HTTP_HDR_URL].b;
#else
  ssize_t uri_len = ctx->http_req->hd[HTTP_HDR_URL].e - ctx->http_req->hd[HTTP_HDR_URL].b;

  const char *question = (const char *)memchr(ctx->http_req->hd[HTTP_HDR_URL].b, '?', uri_len);
  if (question) {
    uri_len = question - ctx->http_req->hd[HTTP_HDR_URL].b;
  }

  const char *uri = ctx->http_req->hd[HTTP_HDR_URL].b;
#endif

  if (cfg->re_exclusion != NULL) {
    LOG(ctx, SLT_Debug, "DataDome test Regex exclusion: '%s' on uri: %s", cfg->re_exclusion_raw, uri);
    if (VRE_exec(cfg->re_exclusion, uri, uri_len, 0, 0, NULL, 0, NULL) >= 0) {
      LOG(ctx, SLT_Debug, "DataDome regex exclusion pass");
      return 0;
    }

    LOG(ctx, SLT_Debug, "DataDome regex exclusion miss");
  }

  if (cfg->re != NULL) {
    LOG(ctx, SLT_Debug, "DataDome test Regex: '%s' on uri: %s", cfg->re_raw, uri);
    if (VRE_exec(cfg->re, uri, uri_len, 0, 0, NULL, 0, NULL) >= 0) {
      LOG(ctx, SLT_Debug, "DataDome regex pass");
      return 1;
    }

    LOG(ctx, SLT_Debug, "DataDome regex miss");
    return 0;
  }

  return 1;
}

static void add_vcl_log_variables(VRT_CTX, struct vmod_data_dome_shield_cfg *cfg, int status, struct timeval *spent)
{
#ifdef VARNISH_VERSION_BEFORE_410
  char *cp;
#else
  const char *cp;
#endif

  if (status > 0) {
    LOG(ctx, SLT_VCL_Log, "DataDome_status: %d", status);
    HTTP_PRINTF_REQ_HDR(ctx, "X-DataDome_module_status: %d", status);
  } else {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_X_DataDome_module_status, &cp)) {
      LOG(ctx, SLT_VCL_Log, "DataDome_status: %s", cp);
    }
  }

  if (spent != NULL) {
    double spent_seconds = (double)spent->tv_sec*1000 + (double) spent->tv_usec/1000;
    LOG(ctx, SLT_VCL_Log, "DataDome_spent_time: %.3f", spent_seconds);
    HTTP_PRINTF_REQ_HDR(ctx, "X-DataDome_module_spent_time: %.3f", spent_seconds);
  } else {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_X_DataDome_module_spent_time, &cp)) {
      LOG(ctx, SLT_VCL_Log, "DataDome_spent_time: %s", cp);
    }
  }
}

VCL_BOOL vmod_is_suitable(VRT_CTX, struct vmod_priv *priv_top)
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  add_vcl_log_variables(ctx, cfg, -1, NULL);
  if (CTX_REQ && CTX_REQ->restarts > 0) {
    LOG(ctx, SLT_Debug, "DataDome request has restarts");

    return 0;
  }

  if (CTX_REQ && CTX_REQ->esi_level > 0) {
    LOG(ctx, SLT_Debug, "DataDome request has esi_level bigger than 0");

    return 0;
  }

  if (cfg->re != NULL || cfg->re_exclusion) {
    if (!vmod_is_regex_matched(ctx, priv_top)) {
      add_vcl_log_variables(ctx, cfg, 700, NULL);
      return 0;
    }
  }

  LOG(ctx, SLT_Debug, "DataDome request is suitable");
  return 1;
}

char to_hex(char code) {
  static char hex[] = "0123456789abcdef";
  return hex[code & 15];
}

const char *url_encode_start(const char *str, ssize_t value_len, ssize_t limit) {
  do {
    str--;
    value_len--;
    limit--;
    if (!isalnum(*str) && *str != '-' && *str != '_' && *str != '.' && *str != '~' && *str != ' ') {
      limit -= 2;
    }
  } while (value_len > 0 && limit > 0);

  return str;
}

ssize_t url_encode(const char *str, char *buf, ssize_t value_len, ssize_t limit) {
  const char *pstr = str;
  char *pbuf = buf;

  while (value_len && *pstr && limit) {
    if (isalnum(*pstr) || *pstr == '-' || *pstr == '_' || *pstr == '.' || *pstr == '~') {
      *pbuf++ = *pstr;
      limit--;
    } else if (*pstr == ' ') {
      *pbuf++ = '+';
      limit--;
    } else {
      if (limit < 3) {
        break;
      }
      *pbuf++ = '%';
      *pbuf++ = to_hex(*pstr >> 4);
      *pbuf++ = to_hex(*pstr & 15);
      limit -= 3;
    }
    pstr++;
    if (value_len > 0) {
      value_len--;
    }
  }

  return pbuf-buf;
}

void add_param(VRT_CTX, struct vmod_data_dome_shield_cfg *cfg, struct vmod_data_dome_shield_status *status,
         const char *name, const char *value, ssize_t value_len, ssize_t limit, unsigned use_url_encode, unsigned from_end)
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
  CHECK_OBJ_NOTNULL(status, VMOD_DATA_DOME_SHIELD_STATUS_MAGIC);

  if (value && value_len < 0) {
    value_len = strlen(value);
  }

  if (value_len == 0) {
    return;
  }

  if (limit > (sizeof(status->body) - status->body_len - 1)) {
    limit = sizeof(status->body) - status->body_len - 1;
  }

  if (from_end && limit > 0) {
    const char* end = value + value_len;

    if (use_url_encode) {
      value = url_encode_start(end, value_len, limit);
      value_len = end - value;
    } else if (value_len > limit) {
      value = end - limit;
      value_len = end - value;
    }

  }

  if (limit > 0 && value_len > limit) {
    value_len = limit;
  }

  LOG(ctx, SLT_Debug, "DataDome add param: name %s, value %p, value_len %zd", name, value, value_len);

  ssize_t max_length = sizeof("&") - 1 + strlen(name) + sizeof("=") - 1 + value_len;

  if (max_length > sizeof(status->body) - status->body_len) {
    return;
  }

  if (status->body_len) {
    status->body[status->body_len] = '&';
    status->body_len++;
  }

  status->body_len +=
    snprintf(status->body + status->body_len, sizeof(status->body) - status->body_len, "%s=", name);

  if (value) {
    if (use_url_encode) {
      status->body_len += url_encode(value, status->body + status->body_len, value_len, limit);
    } else {
      if (value_len > limit) {
        value_len = limit;
      }
      memcpy(status->body + status->body_len, value, value_len);
      status->body_len += value_len;
    }
  }
}


void vmod_data_dome_shield_lookup_client_id(VRT_CTX,
         struct vmod_data_dome_shield_cfg *cfg, struct vmod_data_dome_shield_client_id *client_id)
{
#ifdef VARNISH_VERSION_BEFORE_410
  char *cp, *end;
#else
  char       *end;
  const char *cp;
#endif

  memset(client_id, 0, sizeof(struct vmod_data_dome_shield_client_id));

  if (HTTP_GET_REQ_HDR(ctx, cfg->H_Cookie, &cp)) {
    while (1) {
      cp = strstr((char *)cp, cfg->cookie);
      if (cp == NULL) {
  break;
      }

      cp += strlen(cfg->cookie);

      while (*cp == ' ') {
  cp++;
      }

      if (*cp != '=') {
  cp++;
  continue;
      }

      cp++; //skip =

      while (*cp == ' ') {
  cp++;
      }

      end = strchr(cp, ';');
      if (end == NULL) {
  end = (char *)(cp + strlen(cp));
      }

      client_id->client_id = WS_Alloc(ctx->ws, end - cp + 1);
      AN(client_id->client_id);

      memcpy(client_id->client_id, cp, end - cp);
      client_id->client_id[end - cp] = '\0';

      return;
    }
  }
}

#ifndef VARNISH_VERSION_BEFORE_410
void free_status(void *ptr) {
  struct vmod_data_dome_shield_status *status = (struct vmod_data_dome_shield_status *)ptr;

  status->priv_task->priv = NULL;
  status->priv_task->free = NULL;

  memset(status, 0, sizeof(struct vmod_data_dome_shield_status));

  FREE_OBJ(status);
}
#endif

#ifndef VARNISH_VERSION_BEFORE_500
// from bin/varnishd/http1/cache_http1.h
void V1F_Setup_Fetch(struct vfp_ctx *vfc, struct http_conn *htc);
#endif

#ifdef VARNISH_VERSION_BEFORE_410
VCL_VOID vmod_prepare_request(VRT_CTX, struct vmod_priv *priv_top)
#else
VCL_VOID vmod_prepare_request(VRT_CTX, struct vmod_priv *priv_top, struct vmod_priv *priv_task)
#endif
{
#ifdef VARNISH_VERSION_BEFORE_410
#ifdef VARNISH_VERSION_BEFORE_400
  LOG(ctx, SLT_Debug, "DataDome prepare request, old backend: %s, client: %s:%s",
      ctx->director->name, ctx->addr, ctx->port);
#else
  VSLb(ctx->vsl, SLT_Debug, "DataDome prepare request, old backend: %s, client: %s:%s",
       ctx->req->director_hint->name, ctx->req->sp->client_addr_str, ctx->req->sp->client_port_str);
#endif
#endif

  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  assert(CTX_REQ->restarts == 0);

#ifndef VARNISH_VERSION_BEFORE_410
  AZ(priv_task->priv);
#endif

  struct vmod_data_dome_shield_status *status;
  ALLOC_OBJ(status, VMOD_DATA_DOME_SHIELD_STATUS_MAGIC);
  AN(status);

#ifdef VARNISH_VERSION_BEFORE_410
  set_status(ctx, cfg, status);
#else
  status->priv_task = priv_task;

  priv_task->priv = status;
  priv_task->free = free_status;

  LOG(ctx, SLT_Debug, "DataDome prepare request, old backend: %s", ctx->req->director_hint->vcl_name);
#endif

  memset(&status->tv, 0, sizeof(struct timeval));

  gettimeofday(&status->tv, NULL);

#ifdef VARNISH_VERSION_BEFORE_400
  status->orign_backend = ctx->director;
  ctx->director = cfg->backend;
#else
  status->orign_backend = ctx->req->director_hint;
  ctx->req->director_hint = cfg->backend;
#endif

#ifdef VARNISH_VERSION_BEFORE_410
#ifdef VARNISH_VERSION_BEFORE_400
  status->orign_pipeline_b = ctx->htc->pipeline.b;
  status->orign_pipeline_e = ctx->htc->pipeline.e;
  status->orign_rxbuf_b = ctx->htc->rxbuf.b;
  status->orign_rxbuf_e = ctx->htc->rxbuf.e;
#else
  status->orign_pipeline_b = ctx->req->htc->pipeline.b;
  status->orign_pipeline_e = ctx->req->htc->pipeline.e;
  status->orign_rxbuf_b = ctx->req->htc->rxbuf.b;
  status->orign_rxbuf_e = ctx->req->htc->rxbuf.e;
#endif
#else
  status->orign_pipeline_b = ctx->req->htc->pipeline_b;
  status->orign_pipeline_e = ctx->req->htc->pipeline_e;
  status->orign_rxbuf_b = ctx->req->htc->rxbuf_b;
  status->orign_rxbuf_e = ctx->req->htc->rxbuf_e;
#endif

  vmod_data_dome_shield_lookup_client_id(ctx, cfg, &status->client_id);

  char *p;

  status->body_len = 0;

  add_param(ctx, cfg, status, "Key", cfg->key, -1, -1, 0, 0);

#ifdef VARNISH_VERSION_BEFORE_410
  char *cp;
#else
  const char *cp;
#endif

  if (cfg->send_user_agent) {
    if (HTTP_GET_REQ_HDR(ctx, H_User_Agent, &cp)) {
      add_param(ctx, cfg, status, "UserAgent", cp, -1, 768, 1, 0);
    }
  }

#ifdef VARNISH_VERSION_BEFORE_630
  if (cfg->send_ip) {
#ifdef VARNISH_VERSION_BEFORE_410
#ifdef VARNISH_VERSION_BEFORE_400
    p = ctx->addr;
#else
    p = ctx->req->sp->client_addr_str;
#endif
#else
    AZ(SES_Get_client_ip(ctx->req->sp, &p));
#endif
    add_param(ctx, cfg, status, "IP", p, -1, -1, 0, 0);
  }

  if (cfg->send_port) {
#ifdef VARNISH_VERSION_BEFORE_410
#ifdef VARNISH_VERSION_BEFORE_400
    p = ctx->port;
#else
    p = ctx->req->sp->client_port_str;
#endif
#else
    AZ(SES_Get_client_port(ctx->req->sp, &p));
#endif
    add_param(ctx, cfg, status, "Port", p, -1, -1, 0, 0);
  }
#else
  if (cfg->send_ip) {
    cp = SES_Get_String_Attr(ctx->req->sp, SA_CLIENT_IP);
    AN(cp);
    add_param(ctx, cfg, status, "IP", cp, -1, -1, 0, 0);
  }

  if (cfg->send_port) {
    cp = SES_Get_String_Attr(ctx->req->sp, SA_CLIENT_PORT);
    AN(cp);
    add_param(ctx, cfg, status, "Port", cp, -1, -1, 0, 0);
  }
#endif

  if (cfg->send_ClientID) {
    add_param(ctx, cfg, status, "ClientID", status->client_id.client_id, -1, 128, 1, 0);
  }

  if (cfg->send_host) {
    if (HTTP_GET_REQ_HDR(ctx, H_Host, &cp)) {
      add_param(ctx, cfg, status, "Host", cp, -1, 512, 1, 0);
    }
  }

  if (cfg->send_referer) {
    if (HTTP_GET_REQ_HDR(ctx, H_Referer, &cp)) {
      add_param(ctx, cfg, status, "Referer", cp, -1, 1024, 1, 0);
    }
  }

  if (cfg->send_request) {
    add_param(ctx, cfg, status, "Request", CTX_HTTP_REQ->hd[HTTP_HDR_URL].b, CTX_HTTP_REQ->hd[HTTP_HDR_URL].e - CTX_HTTP_REQ->hd[HTTP_HDR_URL].b, 2048, 1, 0);
  }

  if (cfg->send_protocol) {
    add_param(ctx, cfg, status, "Protocol", cfg->protocol, -1, -1, 0, 0);
  }

  if (cfg->send_method) {
    add_param(ctx, cfg, status, "Method", CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].b, CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].e - CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].b, -1, 0, 0);
  }

  if (cfg->send_cookies_len) {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_Cookie, &cp)) {
      const char t[sizeof("2147483647")];
      ssize_t len = snprintf((char *)t, sizeof(t), "%d", (int)strlen(cp));
      add_param(ctx, cfg, status, "CookiesLen", t, len, -1, 1, 0);
    } else {
      add_param(ctx, cfg, status, "CookiesLen", "0", -1, -1, 0, 0);
    }
  }

  if (cfg->send_time_request) {
    char curr_time[64];

    ssize_t l = snprintf(curr_time, sizeof(curr_time), "%li%06li", (long)status->tv.tv_sec, (long)status->tv.tv_usec);

    add_param(ctx, cfg, status, "TimeRequest", curr_time, l, -1, 0, 0);
  }

  if (cfg->send_server_hostname) {
    if (HTTP_GET_REQ_HDR(ctx, H_Host, &cp)) {
      add_param(ctx, cfg, status, "ServerHostname", cp, -1, 512, 1, 0);
    }
  }

  if (cfg->send_module_name) {
    add_param(ctx, cfg, status, "RequestModuleName", MODULE_NAME, -1, -1, 0, 0);
  }

  if (cfg->send_module_version) {
    add_param(ctx, cfg, status, "ModuleVersion", VERSION, -1, -1, 0, 0);
  }

  if (cfg->send_post_param_len) {
    if (HTTP_GET_REQ_HDR(ctx, H_Content_Length, &cp)) {
      add_param(ctx, cfg, status, "PostParamLen", cp, -1, -1, 0, 0);
    } else {
      add_param(ctx, cfg, status, "PostParamLen", "0", -1, -1, 0, 0);
    }
  }

  if (cfg->send_server_name) {
    add_param(ctx, cfg, status, "ServerName", cfg->host_name, -1, 512, 0, 0);
  }

  if (cfg->send_x_forward_for_ip) {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_X_Forwarded_For, &cp)) {
#ifdef VARNISH_VERSION_BEFORE_400
      const char *comma = NULL;
#else
      const char *comma = strrchr(cp, ',');
#endif
      add_param(ctx, cfg, status, "XForwaredForIP", cp, comma ? comma - cp : -1, 512, 1, 1);
    }
  }

  if (cfg->send_headers_list) {
    txt *h;
    ssize_t headers_len = 0;
    char *t, *tt;
    for (h = &CTX_HTTP_REQ->hd[HTTP_HDR_FIRST]; (unsigned char *)h != CTX_HTTP_REQ->hdf; h++) {
        if (h->b) {
            if (headers_len > 0) {
                headers_len += 1; // ','
            }
            for (t = (char *)h->b; t != h->e && *t != ':'; t++);
            headers_len += t - h->b;
        }
    }

    p = WS_Alloc(ctx->ws, headers_len);
    if (p == NULL) {
        return;
    }

    tt = p;
    headers_len = 0;

    for (h = &CTX_HTTP_REQ->hd[HTTP_HDR_FIRST]; (unsigned char *)h != CTX_HTTP_REQ->hdf; h++) {
        if (h->b) {
            if (headers_len > 0) {
                headers_len += 1; // ','
                *tt = ',';
                tt++;
            }
            for (t = (char *)h->b; t != h->e && *t != ':'; t++) {
                *tt = tolower(*t);
                tt++;
            }
            headers_len += t - h->b;
        }
    }

    add_param(ctx, cfg, status, "HeadersList", p, headers_len, 512, 1, 0);
  }

  if (cfg->send_authorization_len) {
    if (HTTP_GET_REQ_HDR(ctx, H_Authorization, &cp)) {
      const char t[sizeof("2147483647")];
      ssize_t len = snprintf((char *)t, sizeof(t), "%d", (int)strlen(cp));
      add_param(ctx, cfg, status, "AuthorizationLen", t, len, -1, 0, 0);
    } else {
      add_param(ctx, cfg, status, "AuthorizationLen", "0", -1, -1, 0, 0);
    }
  }

  if (cfg->send_x_requested_with) {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_X_Requested_With, &cp)) {
      add_param(ctx, cfg, status, "X-Requested-With", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_origin) {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_Origin, &cp)) {
      add_param(ctx, cfg, status, "Origin", cp, -1, 512, 1, 0);
    }
  }

  if (cfg->send_connection) {
    if (HTTP_GET_REQ_HDR(ctx, H_Connection, &cp)) {
      add_param(ctx, cfg, status, "Connection", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_pragma) {
    if (HTTP_GET_REQ_HDR(ctx, H_Pragma, &cp)) {
      add_param(ctx, cfg, status, "Pragma", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_cache_control) {
    if (HTTP_GET_REQ_HDR(ctx, H_Cache_Control, &cp)) {
      add_param(ctx, cfg, status, "CacheControl", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_content_type) {
    if (HTTP_GET_REQ_HDR(ctx, H_Content_Type, &cp)) {
      add_param(ctx, cfg, status, "ContentType", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_from) {
    if (HTTP_GET_REQ_HDR(ctx, H_From, &cp)) {
      add_param(ctx, cfg, status, "From", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_x_real_ip) {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_X_Real_IP, &cp)) {
      add_param(ctx, cfg, status, "X-Real-IP", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_via) {
    if (HTTP_GET_REQ_HDR(ctx, H_Via, &cp)) {
      add_param(ctx, cfg, status, "Via", cp, -1, 256, 1, 0);
    }
  }

  if (cfg->send_true_client_ip) {
    if (HTTP_GET_REQ_HDR(ctx, cfg->H_True_Client_IP, &cp)) {
      add_param(ctx, cfg, status, "TrueClientIP", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_accept) {
    if (HTTP_GET_REQ_HDR(ctx, H_Accept, &cp)) {
      add_param(ctx, cfg, status, "Accept", cp, -1, 512, 1, 0);
    }
  }

  if (cfg->send_accept_charset) {
    if (HTTP_GET_REQ_HDR(ctx, H_Accept_Charset, &cp)) {
      add_param(ctx, cfg, status, "AcceptCharset", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_accept_encoding) {
    if (HTTP_GET_REQ_HDR(ctx, H_Accept_Encoding, &cp)) {
      add_param(ctx, cfg, status, "AcceptEncoding", cp, -1, 128, 1, 0);
    }
  }

  if (cfg->send_accept_language) {
    if (HTTP_GET_REQ_HDR(ctx, H_Accept_Language, &cp)) {
      add_param(ctx, cfg, status, "AcceptLanguage", cp, -1, 256, 1, 0);
    }
  }

#ifndef VARNISH_VERSION_BEFORE_400
  status->orign_req_body_status = ctx->req->req_body_status;
#ifdef VARNISH_VERSION_BEFORE_640
#ifdef VARNISH_VERSION_BEFORE_410
  ctx->req->req_body_status = REQ_BODY_PRESENT;
#else  // 410
  ctx->req->req_body_status = REQ_BODY_WITH_LEN;
#endif // 410
#else
  ctx->req->req_body_status = BS_LENGTH;
#endif // 640
#endif // 499

#ifdef VARNISH_VERSION_BEFORE_400
  status->orign_body_status = ctx->wrk->body_status;
  ctx->wrk->body_status = BS_LENGTH;
#else
  status->orign_body_status = ctx->req->htc->body_status;
  ctx->req->htc->body_status = BS_LENGTH;
#endif

  if (status->body_len > MAX_PARAM_DATALEN) {
    status->body_len = MAX_PARAM_DATALEN;
  }

#ifdef VARNISH_VERSION_BEFORE_400
  ctx->htc->pipeline.b = status->body;
  ctx->htc->pipeline.e = status->body + status->body_len;
  ctx->htc->rxbuf.b = ctx->htc->pipeline.b;
  ctx->htc->rxbuf.e = ctx->htc->pipeline.e;
#else
#ifdef VARNISH_VERSION_BEFORE_410
  ctx->req->htc->pipeline.b = status->body;
  ctx->req->htc->pipeline.e = status->body + status->body_len;
  ctx->req->htc->rxbuf.b = ctx->req->htc->pipeline.b;
  ctx->req->htc->rxbuf.e = ctx->req->htc->pipeline.e;
#else
  ctx->req->htc->pipeline_b = status->body;
  ctx->req->htc->pipeline_e = status->body + status->body_len;
  ctx->req->htc->rxbuf_b = ctx->req->htc->pipeline_b;
  ctx->req->htc->rxbuf_e = ctx->req->htc->pipeline_e;
#endif
#endif

#ifdef VARNISH_VERSION_BEFORE_400
  status->orign_hd = ctx->http->hd;
  status->orign_shd = ctx->http->shd;
  status->orign_hdf = ctx->http->hdf;
  status->orign_nhd = ctx->http->nhd;
#else
#ifdef VARNISH_VERSION_BEFORE_410
  status->orign_content_length = ctx->req->req_bodybytes;
  ctx->req->req_bodybytes = status->body_len;
#else
  status->orign_content_length = ctx->req->htc->content_length;
  ctx->req->htc->content_length = status->body_len;
#endif
#endif

#ifndef VARNISH_VERSION_BEFORE_500
  // we use body but request mayn't have the cached body
  // so, it will crash since
  // https://github.com/varnishcache/varnish-cache/commit/e8cae033eab21cce6ca4ca8f94cf4cd693d639a3
#ifdef VARNISH_VERSION_BEFORE_510
  *status->orign_vfc = *ctx->req->htc->vfc;
  memset(ctx->req->htc->vfc, 0, sizeof(struct vfp_ctx));
#else
  *status->orign_vfc = *ctx->req->vfc;
  memset(ctx->req->vfc, 0, sizeof(struct vfp_ctx));
#endif

#ifdef VARNISH_VERSION_BEFORE_600
#ifdef VARNISH_VERSION_BEFORE_510
  VFP_Setup(ctx->req->htc->vfc);
  ctx->req->htc->vfc->http = ctx->req->http;
  ctx->req->htc->vfc->wrk = ctx->req->wrk;
#else
  VFP_Setup(ctx->req->vfc);
  ctx->req->vfc->http = ctx->req->http;
  ctx->req->vfc->wrk = ctx->req->wrk;
#endif // BEFORE_510
#else
  VFP_Setup(ctx->req->vfc, ctx->req->wrk);
  ctx->req->vfc->resp = ctx->req->http;
#endif

#ifdef VARNISH_VERSION_BEFORE_510
  V1F_Setup_Fetch(ctx->req->htc->vfc, ctx->req->htc);
#else
  V1F_Setup_Fetch(ctx->req->vfc, ctx->req->htc);
#endif

#endif // AFTER_500

#ifdef VARNISH_VERSION_BEFORE_400
  ssize_t new_http_len = HTTP_estimate(ctx->http->shd);
#else
  status->orign_http_req = CTX_HTTP_REQ;
  status->orign_req_http = ctx->req->http;

  ssize_t new_http_len = HTTP_estimate(status->orign_http_req->shd);
#endif

  void *new_http = WS_Alloc(ctx->ws, new_http_len);
  AN(new_http);

#ifdef VARNISH_VERSION_BEFORE_400
  memset(new_http, 0, new_http_len);

  ctx->http->hd = (void *)((struct http *)new_http + 1);
  ctx->http->hdf = (void*)(ctx->http->hd + ctx->http->shd);

  ctx->http->nhd = HTTP_HDR_FIRST;
#else
#ifdef VARNISH_VERSION_BEFORE_512
  *((struct http **) &CTX_HTTP_REQ) = HTTP_create(new_http, status->orign_http_req->shd);
#else
  *((struct http **) &CTX_HTTP_REQ) = HTTP_create(new_http, status->orign_http_req->shd, new_http_len);
#endif
  ctx->req->http = ctx->http_req;

  HTTP_Setup(ctx->http_req, status->orign_http_req->ws, status->orign_http_req->vsl, status->orign_http_req->logtag);
#endif

  CTX_HTTP_REQ->hd[HTTP_HDR_URL].b = WS_Alloc(ctx->ws, cfg->validate_url.e - cfg->validate_url.b);
  AN(CTX_HTTP_REQ->hd[HTTP_HDR_URL].b);
  memcpy((void *)CTX_HTTP_REQ->hd[HTTP_HDR_URL].b, cfg->validate_url.b, cfg->validate_url.e - cfg->validate_url.b);
  CTX_HTTP_REQ->hd[HTTP_HDR_URL].e = CTX_HTTP_REQ->hd[HTTP_HDR_URL].b + (cfg->validate_url.e - cfg->validate_url.b);

  CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].b = WS_Alloc(ctx->ws, cfg->validate_method.e - cfg->validate_method.b);
  AN(CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].b);
  memcpy((void *)CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].b, cfg->validate_method.b, cfg->validate_method.e - cfg->validate_method.b);
  CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].e = CTX_HTTP_REQ->hd[HTTP_HDR_METHOD].b + (cfg->validate_method.e - cfg->validate_method.b);

  CTX_HTTP_REQ->hd[HTTP_HDR_PROTO].b = WS_Alloc(ctx->ws, cfg->validate_proto.e - cfg->validate_proto.b);
  AN(CTX_HTTP_REQ->hd[HTTP_HDR_PROTO].b);
  memcpy((void *)CTX_HTTP_REQ->hd[HTTP_HDR_PROTO].b, cfg->validate_proto.b, cfg->validate_proto.e - cfg->validate_proto.b);
  CTX_HTTP_REQ->hd[HTTP_HDR_PROTO].e = CTX_HTTP_REQ->hd[HTTP_HDR_PROTO].b + (cfg->validate_proto.e - cfg->validate_proto.b);

  status->headers_restored = 0;

  HTTP_PRINTF_REQ_HDR(ctx, "Content-Length: %zd", status->body_len);
  HTTP_SET_REQ_HDR(ctx, "Content-Type: application/x-www-form-urlencoded");
}


void execute_x_datadome_headers(VRT_CTX, const char* x_datadome_header, struct http *from, struct http *to, int cleanup, int copy_xddh)
{
#ifdef VARNISH_VERSION_BEFORE_410
  char *x_datadome_headers;
#else
  const char *x_datadome_headers;
#endif
  char *p, *key;
  unsigned u, v;

  u = WS_ReserveAll(ctx->ws);
  if (u < 255) {
    WS_Release(ctx->ws, 0);
    return;
  }
  p = ctx->ws->f;

  ssize_t x_datadome_header_len = strlen(x_datadome_header) + 1; // with \0

  memcpy(p + 1, x_datadome_header, x_datadome_header_len);
  v = x_datadome_header_len;
  p[0] = (char) (v - 1);

  WS_Release(ctx->ws, v + 1);

  if (!http_GetHdr(from, p, &x_datadome_headers)) {
    return;
  }

  if (cleanup) {
    http_Unset(from, p);
  }

#ifdef VARNISH_VERSION_BEFORE_400
  if (to != NULL && ctx->wrk->resp != to && copy_xddh) {
#else
  if (to != NULL && ctx->http_resp != to && copy_xddh) {
#endif
      http_Unset(to, p);
      u = WS_ReserveAll(ctx->ws);
      v = 0;
      if (u > 1) {
          p = ctx->ws->f;
          v = 1 + snprintf(p, u - 1, "%s %s", x_datadome_header, x_datadome_headers);
#ifdef VARNISH_VERSION_BEFORE_400
          http_SetHeader(ctx->wrk, ctx->wrk->htc->fd, to, p);
#else
          http_SetHeader(to, p);
#endif
      }
      WS_Release(ctx->ws, v);
  }

  while (strlen(x_datadome_headers) > 0) {
    u = WS_ReserveAll(ctx->ws);
    if (u < 255) {
        WS_Release(ctx->ws, 0);
        return;
    }
    p = ctx->ws->f;

#ifdef VARNISH_VERSION_BEFORE_410
    char *end = strstr(x_datadome_headers, " ");
    if (end == NULL) {
      end = x_datadome_headers + strlen(x_datadome_headers);
    }
#else
    const char *end = strstr(x_datadome_headers, " ");
    if (end == NULL) {
        end = (const char *) x_datadome_headers + strlen(x_datadome_headers);
    }
#endif

    memcpy(p + 1, x_datadome_headers, end - x_datadome_headers);
    p[1 + end - x_datadome_headers] = ':';
    p[1 + end - x_datadome_headers + 1] = '\0';

    v = 1 + strlen(p + 1);
    p[0] = (char) (v - 1);

    WS_Release(ctx->ws, v + 1);

#ifdef VARNISH_VERSION_BEFORE_410
    char *header;
#else
    const char *header;
#endif

    if (http_GetHdr(from, p, &header)) {
      if (cleanup) {
        http_Unset(from, p);
      }

      if (to != NULL) {
        if (end - x_datadome_headers != sizeof("Set-Cookie") - 1
         ||  strncasecmp("Set-Cookie", x_datadome_headers, sizeof("Set-Cookie") -1 )) {
          http_Unset(to, p);
        }

        key = p + 1;

        u = WS_ReserveAll(ctx->ws);
        v = 0;
        if (u > 1) {
            p = ctx->ws->f;
            v = 1 + snprintf(p, u - 1, "%s %s", key, header);
#ifdef VARNISH_VERSION_BEFORE_400
            http_SetHeader(ctx->wrk, ctx->wrk->htc->fd, to, p);
#else
            http_SetHeader(to, p);
#endif
        }
        WS_Release(ctx->ws, v);
      }
    }

    x_datadome_headers = end;
    while (*x_datadome_headers == ' ') {
        x_datadome_headers++;
    }
  }

  // cleanup all X-DataDome headers
  if (cleanup) {
  uint16_t u, v;
  for (v = u = HTTP_HDR_FIRST; u < from->nhd; u++) {
    if (from->hd[u].b == NULL)
      continue;
    if (from->hd[u].e - from->hd[u].b >= sizeof("X-DataDome") - 1 &&
        !strncasecmp(from->hd[u].b, "X-DataDome", sizeof("X-DataDome") - 1)) {
        continue;
    }
    if (v != u) {
      memcpy(&from->hd[v], &from->hd[u], sizeof *from->hd);
      memcpy(&from->hdf[v], &from->hdf[u], sizeof *from->hdf);
    }
    v++;
  }
  from->nhd = v;
  }
}


VCL_VOID vmod_prepare_request_backend(VRT_CTX, struct vmod_priv *priv_top)
{
  vmod_prepare_backends_request(ctx, priv_top);
}


VCL_VOID vmod_prepare_backends_request(VRT_CTX, struct vmod_priv *priv_top)
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

  // force call cleanup backend's request
  vmod_cleanup_backends_request(ctx, priv_top);

#ifdef VARNISH_VERSION_BEFORE_400
  if (ctx->vbc == NULL || vdi_get_backend_if_simple(cfg->backend) != ctx->vbc->backend) {
#else
#ifdef VARNISH_VERSION_BEFORE_410
  if (cfg->backend != ctx->bo->director) {
#else
  if (cfg->backend != ctx->bo->director_req) {
#endif
#endif
    return;
  }

#ifdef VARNISH_VERSION_BEFORE_400
  http_Unset(ctx->wrk->bereq, H_Connection);
  http_SetHeader(ctx->wrk, ctx->vbc->fd, ctx->wrk->bereq, "Connection: keep-alive");
#else
  http_Unset(ctx->http_bereq, H_Connection);
  http_SetHeader(ctx->http_bereq, "Connection: keep-alive");
#endif
}


VCL_VOID vmod_cleanup_backends_request(VRT_CTX, struct vmod_priv *priv_top)
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_400
  struct http *temp = ctx->http;
#else
  // someone retry to connect to backend, it was cleaned up at the first try, ignore it
  if (ctx->bo != NULL && ctx->bo->retries > 0) {
    return;
  }

  // we haven't access to ctx->http_req here, so, use a temporary request to keep the headers
  struct http *temp;
  char *p;
  unsigned u, v;

  v = HTTP_estimate(ctx->http_bereq->shd);

  u = WS_ReserveAll(ctx->ws);
  if (u < v) {
    WS_Release(ctx->ws, 0);
    return;
  }
  p = ctx->ws->f;
#ifdef VARNISH_VERSION_BEFORE_512
  temp = HTTP_create(p, ctx->http_bereq->shd);
#else
  temp = HTTP_create(p, ctx->http_bereq->shd, v);
#endif
  WS_Release(ctx->ws, v);

  execute_x_datadome_headers(ctx, "X-DataDome-request-headers:", ctx->http_bereq, temp, 0, 1);
#endif

  // cleanup X-DataDome-Headers
#ifdef VARNISH_VERSION_BEFORE_400
  execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->wrk->bereq, NULL, 1, 0);
#else
  execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->http_bereq, NULL, 1, 0);
#endif

  // copy request headers to backend request and cleanup it
#ifdef VARNISH_VERSION_BEFORE_400
  execute_x_datadome_headers(ctx, "X-DataDome-request-headers:", temp, ctx->wrk->bereq, 0, 0);
#else
  execute_x_datadome_headers(ctx, "X-DataDome-request-headers:", temp, ctx->http_bereq, 0, 0);
#endif
}

#ifdef VARNISH_VERSION_BEFORE_410
 VCL_BOOL vmod_has_prepared_request(VRT_CTX, struct vmod_priv *priv_top)
#else
VCL_BOOL vmod_has_prepared_request(VRT_CTX, struct vmod_priv *priv_top, struct vmod_priv *priv_task)
#endif
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_410
  struct vmod_data_dome_shield_status *status = get_status(ctx, cfg);
#else
  struct vmod_data_dome_shield_status *status = priv_task->priv;
#endif

  LOG(ctx, SLT_Debug, "DataDome has prepared request %d", status != NULL);

  if (status != NULL && status->magic == VMOD_DATA_DOME_SHIELD_STATUS_MAGIC) {
    return 1;
  } else {
#ifdef VARNISH_VERSION_BEFORE_400
    execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->http, ctx->wrk->resp, 0, 0);
#else
    execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->http_req, ctx->http_resp, 0, 0);
#endif
    return 0;
  }
}

VCL_BOOL vmod_is_datadome_call(VRT_CTX, struct vmod_priv *priv_top)
{
  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_400
  return cfg->backend == ctx->director;
#else
#ifdef VARNISH_VERSION_BEFORE_410
  return cfg->backend == ctx->bo->director;
#else
  return cfg->backend == ctx->bo->director_req;
#endif
#endif
}


void restore_headers(VRT_CTX, struct vmod_data_dome_shield_status *status) {
  if (status->headers_restored) {
    return;
  }

#ifdef VARNISH_VERSION_BEFORE_400
  ctx->http->hd = status->orign_hd;
  ctx->http->shd = status->orign_shd;
  ctx->http->hdf = status->orign_hdf;
  ctx->http->nhd = status->orign_nhd;
#else
  *((struct http **) &ctx->http_req) = status->orign_http_req;
  ctx->req->http = status->orign_req_http;
#endif

  status->headers_restored = 1;
}

#ifdef VARNISH_VERSION_BEFORE_410
VCL_BOOL vmod_is_valid(VRT_CTX, struct vmod_priv *priv_top)
#else
VCL_BOOL vmod_is_valid(VRT_CTX, struct vmod_priv *priv_top, struct vmod_priv *priv_task)
#endif
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_410
  struct vmod_data_dome_shield_status *status = get_status(ctx, cfg);
#else
  struct vmod_data_dome_shield_status *status = priv_task->priv;
#endif
  CHECK_OBJ_NOTNULL(status, VMOD_DATA_DOME_SHIELD_STATUS_MAGIC);

  int is_valid = 0;

  struct timeval tv, res;

  memset(&tv, 0, sizeof(struct timeval));
  memset(&res, 0, sizeof(struct timeval));

  gettimeofday(&tv, NULL);

  timersub(&tv, &status->tv, &res);

#ifdef VARNISH_VERSION_BEFORE_410
  vmod_restore_request(ctx, priv_top);
#else
  vmod_restore_request(ctx, priv_top, priv_task);
#endif

#ifdef VARNISH_VERSION_BEFORE_410
  char *x_datadome_response, *cp;
#else
  const char *x_datadome_response, *cp;
#endif
  char *p;
  unsigned u, v;

  u = WS_ReserveAll(ctx->ws);
  if (u < 255) {
    WS_Release(ctx->ws, 0);
    return 1;
  }
  p = ctx->ws->f;

  memcpy(p + 1, "X-DataDomeResponse:", sizeof("X-DataDomeResponse:"));
  v = 1 + strlen(p + 1);
  p[0] = (char) (v - 1);
  p[v + 1] = '\0';

  WS_Release(ctx->ws, v + 1);

  if (!HTTP_GET_RESP_HDR(ctx, p, &x_datadome_response)) {
    LOG(ctx, SLT_Error, "DataDome api return response without X-DataDomeResponse");
    add_vcl_log_variables(ctx, cfg, 704, &res);
    return 1;
  }

#ifdef VARNISH_VERSION_BEFORE_400
  int response_status = ctx->wrk->resp->status;
#else
  int response_status = ctx->http_resp->status;
#endif
  if (atoi(x_datadome_response) != response_status) {
    LOG(ctx, SLT_Error, "DataDome api returns response where X-DataDomeResponse(%s) != status(%d)",
         x_datadome_response, response_status);
    add_vcl_log_variables(ctx, cfg, 704, &res);
    return 1;
  }

  switch (response_status) {
  case 301:
  case 302:
  case 401:
  case 403:
    is_valid = 0;
    break;

  case 200:
    is_valid = 1;
    break;

  default:
    is_valid = 1;

    LOG(ctx, SLT_Error, "DataDome api return unexpectable status: %d", response_status);
  }

  LOG(ctx, SLT_Debug, "DataDome is valid %d", is_valid);

#ifdef VARNISH_VERSION_BEFORE_400
  execute_x_datadome_headers(ctx, "X-DataDome-request-headers:", ctx->wrk->resp, ctx->http, 0, 1);
  execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->wrk->resp, ctx->http, 1, 1);
  execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->http, ctx->wrk->resp, 0, 0);
#else
  execute_x_datadome_headers(ctx, "X-DataDome-request-headers:", ctx->http_resp, ctx->http_req, 0, 1);
  execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->http_resp, ctx->http_req, 1, 1);
  execute_x_datadome_headers(ctx, "X-DataDome-headers:", ctx->http_req, ctx->http_resp, 0, 0);
#endif

  if (!is_valid && cfg->is_valid_force_close_banned_request
      && (HTTP_IS_REQ_HDR(ctx, H_Transfer_Encoding, "chunked")
          || (HTTP_GET_REQ_HDR(ctx, H_Content_Length, &cp) && !HTTP_IS_REQ_HDR(ctx, H_Content_Length, "0")))) {
#ifdef VARNISH_VERSION_BEFORE_400
    http_Unset(ctx->wrk->resp, H_Connection);
#endif
    HTTP_SET_RESP_HDR(ctx, "Connection: close");
#ifdef VARNISH_VERSION_BEFORE_400
    ctx->doclose = "close";
#else
#ifdef VARNISH_VERSION_BEFORE_410
      ctx->req->doclose = SC_REQ_CLOSE;
#else
      ctx->req->htc->doclose = SC_REQ_CLOSE;
#endif
#endif
  }

  add_vcl_log_variables(ctx, cfg, response_status, &res);
  return is_valid;
}

VCL_VOID vmod_is_valid_force_close_banned_request(VRT_CTX, struct vmod_priv *priv_top, VCL_BOOL value)
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

    cfg->is_valid_force_close_banned_request = value;
}

#ifndef VARNISH_VERSION_BEFORE_620
// I haven't got any easy way to include cache/cache_objhead.h
void HSH_DerefBoc(struct worker *wrk, struct objcore *);
#endif

#ifdef VARNISH_VERSION_BEFORE_410
VCL_VOID vmod_restore_request(VRT_CTX, struct vmod_priv *priv_top)
#else
VCL_VOID vmod_restore_request(VRT_CTX, struct vmod_priv *priv_top, struct vmod_priv *priv_task)
#endif
{
  CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

  struct vmod_data_dome_shield_cfg *cfg = priv_top->priv;
  CHECK_OBJ_NOTNULL(cfg, VMOD_DATA_DOME_SHIELD_CFG_MAGIC);

#ifdef VARNISH_VERSION_BEFORE_410
  struct vmod_data_dome_shield_status *status = get_status(ctx, cfg);
#else
  struct vmod_data_dome_shield_status *status = priv_task->priv;
#endif
  if (status == NULL) {
    return;
  }
  CHECK_OBJ_NOTNULL(status, VMOD_DATA_DOME_SHIELD_STATUS_MAGIC);

  LOG(ctx, SLT_Debug, "DataDome restore request");

#ifdef VARNISH_VERSION_BEFORE_400
  ctx->director = status->orign_backend;
  ctx->wrk->body_status = status->orign_body_status;
  ctx->htc->pipeline.b = status->orign_pipeline_b;
  ctx->htc->pipeline.e = status->orign_pipeline_e;
  ctx->htc->rxbuf.b = status->orign_rxbuf_b;
  ctx->htc->rxbuf.e = status->orign_rxbuf_e;
#else
  ctx->req->director_hint = status->orign_backend;
  ctx->req->req_body_status = status->orign_req_body_status;
  ctx->req->htc->body_status = status->orign_body_status;
#ifdef VARNISH_VERSION_BEFORE_410
  ctx->req->htc->pipeline.b = status->orign_pipeline_b;
  ctx->req->htc->pipeline.e = status->orign_pipeline_e;
  ctx->req->htc->rxbuf.b = status->orign_rxbuf_b;
  ctx->req->htc->rxbuf.e = status->orign_rxbuf_e;
  ctx->req->req_bodybytes = status->orign_content_length;
#else
  ctx->req->htc->pipeline_b = status->orign_pipeline_b;
  ctx->req->htc->pipeline_e = status->orign_pipeline_e;
  ctx->req->htc->rxbuf_b = status->orign_rxbuf_b;
  ctx->req->htc->rxbuf_e = status->orign_rxbuf_e;
  ctx->req->htc->content_length = status->orign_content_length;
#endif

  // re-read body
#ifdef VARNISH_VERSION_BEFORE_620
#ifdef VARNISH_VERSION_BEFORE_410
  ctx->req->h1.bytes_done = 0;
#else
  ctx->req->req_bodybytes = 0;
#endif
#else
  if (ctx->req->body_oc) {
    ObjTrimStore(ctx->req->wrk, ctx->req->body_oc);
    AZ(ObjSetU64(ctx->req->wrk, ctx->req->body_oc, OA_LEN, 0));
    HSH_DerefBoc(ctx->req->wrk, ctx->req->body_oc);
  }
#endif
#endif

#ifndef VARNISH_VERSION_BEFORE_500
#ifdef VARNISH_VERSION_BEFORE_510
  VFP_Close(ctx->req->htc->vfc);
  *ctx->req->htc->vfc = *status->orign_vfc;
#else
  VFP_Close(ctx->req->vfc);
  *ctx->req->vfc = *status->orign_vfc;
#endif
#endif // AFTER_500

  restore_headers(ctx, status);

#ifdef VARNISH_VERSION_BEFORE_410
  remove_status(ctx, cfg, status);
  FREE_OBJ(status);
#else
  free_status(status);
  priv_task->priv = NULL;
#endif
}
